import os
import errno
import socket
import signal
import random
import selectors
import argparse
from bidict import bidict
from collections import deque
from functools import partial
from itertools import chain
from rikka.config import Config, ConfigAttribute
from rikka.exceptions import ConfigMissing
from rikka.logger import logger, name2level
from rikka.protocol import Protocol, PKGBuilder, BUF_SIZE, \
    sentinel
from rikka.utils import parse_netloc, set_non_blocking, \
    format_addr, create_listening_sock


from argparse import Namespace
from socket import socket as socket_t
from typing import List, Optional

POS = 0  # expose to tunnel
NEG = 1  # tunnel to expose


class Server:

    tunnel_addr = ConfigAttribute('tunnel', parse_netloc)
    expose_addr = ConfigAttribute('bind', parse_netloc)

    def __init__(self, pkgbuilder: PKGBuilder, config: Config) -> None:
        self._ready: deque = deque()  # task queue
        self._config = config
        self._stopping = False
        self._pkgbuilder = pkgbuilder
        self._sel = selectors.DefaultSelector()

        self.tunnel_pool: deque = deque()
        self.work_pool = bidict()

        # reinstall signal
        self.init_signal()
        # select timeout
        self.reset_timeout()

        self.start_listen()

    @property
    def config(self) -> Config:
        return self._config

    def start_listen(self) -> None:
        """listen expose conn and tunnel conn"""
        self.expose_sock = create_listening_sock(self.expose_addr)
        self._sel.register(self.expose_sock, selectors.EVENT_READ,
                           self.accept_expose)

        self.tunnel_sock = create_listening_sock(self.tunnel_addr)
        self._sel.register(self.tunnel_sock, selectors.EVENT_READ,
                           self.accept_tunnel)

        self._listening_sock = [
            self.expose_sock,
            self.tunnel_sock,
        ]

    def next_timeout(self) -> None:
        """binary exponential backoff"""
        self._timeout_count += 1
        upper_bound = (2 ** min(self._timeout_count, 7)) - 1
        self._timeout = random.randint(1, upper_bound)

    def reset_timeout(self) -> None:
        """reset timeout to initial value"""
        self._timeout: int = 1  # mypy need this variable type
        self._timeout_count: int = 0

    def accept_expose(self, expose_sock: socket_t, mask: int) -> None:
        """accept user connection"""
        conn, addr = expose_sock.accept()
        conn.setblocking(False)

        self._sel.register(conn, selectors.EVENT_READ, self.prepare_transfer)

        logger.info(f'accept user connection from {format_addr(addr)}')

    def accept_tunnel(self, tunnel_sock: socket_t, mask: int) -> None:
        """accept tunnel connection"""
        conn, addr = tunnel_sock.accept()
        conn.setblocking(False)

        self.tunnel_pool.append(conn)

        logger.info(f'accept tunnel connection from {format_addr(addr)}, '
                    f'poolsize is {len(self.tunnel_pool)}')

    def prepare_transfer(self, expose_conn: socket_t, mask: int) -> None:
        tunnel_conn = self.find_available_tunnel()
        if tunnel_conn is None:  # non-available tunnel_conn
            self._sel.unregister(expose_conn)
            # delay
            self._ready.append(
                lambda: self._sel.register(
                    expose_conn, selectors.EVENT_READ,
                    self.prepare_transfer
                )
            )
            return
        self.work_pool[expose_conn] = tunnel_conn
        buf: List[deque] = [deque(), deque()]
        self._sel.register(tunnel_conn,
                           selectors.EVENT_WRITE | selectors.EVENT_READ,
                           partial(self.dispatch_tunnel, buf=buf))
        self._sel.modify(expose_conn,
                         selectors.EVENT_WRITE | selectors.EVENT_READ,
                         partial(self.dispatch_expose, buf=buf))

    def dispatch_tunnel(self, conn: socket_t,
                        mask: int, buf: List[deque]) -> None:
        """schedule tunnel events"""
        if mask & selectors.EVENT_WRITE:
            self.send_to_tunnel(conn, mask, buf)
        if mask & selectors.EVENT_READ:
            self.transfer_from_tunnel(conn, mask, buf)

    def dispatch_expose(self, conn: socket_t,
                        mask: int, buf: List[deque]) -> None:
        """schedule expose events"""
        if mask & selectors.EVENT_WRITE:
            self.send_to_expose(conn, mask, buf)
        if mask & selectors.EVENT_READ:
            self.transfer_from_expose(conn, mask, buf)

    def transfer_from_expose(self, r_conn: socket_t,
                             mask: int, buf: List[deque]) -> None:
        """receive data from expose and store in buffer"""
        w_conn = self.work_pool.get(r_conn)
        if w_conn is None:
            self._sel.unregister(r_conn)
            r_conn.close()
            return

        data = b''
        need_close = False

        try:
            data = r_conn.recv(BUF_SIZE)  # Connection may be close
        except ConnectionError:
            need_close = True

        if data == b'' or need_close:
            try:
                peer = r_conn.getpeername()
                logger.info(f'closing user connection from {format_addr(peer)}')  # noqa
            except OSError as e:
                logger.warn(e)
            self._sel.unregister(r_conn)
            r_conn.close()
            buf[POS].append(sentinel)
            del self.work_pool[r_conn]
            return

        buf[POS].append(data)

    def transfer_from_tunnel(self, r_conn: socket_t,
                             mask: int, buf: List[deque]) -> None:
        """receive data from tunnel and store in buffer"""
        w_conn = self.work_pool.inv.get(r_conn)
        if w_conn is None:
            self._sel.unregister(r_conn)
            r_conn.close()
            return

        data = b''
        need_close = False

        try:
            data = r_conn.recv(BUF_SIZE)
        except ConnectionError:
            need_close = True

        if data == b'' or need_close:
            try:
                peer = r_conn.getpeername()
                logger.info(f'closing tunnel connection from {format_addr(peer)}')  # noqa
            except OSError as e:
                logger.warn(e)
            self._sel.unregister(r_conn)
            r_conn.close()
            buf[NEG].append(sentinel)
            del self.work_pool.inv[r_conn]
            return

        buf[NEG].append(data)

    def send_to_tunnel(self, w_conn: socket_t,
                       mask: int, buf: List[deque]) -> None:
        """send buffer data to tunnel"""
        if not len(buf[POS]):
            return
        try:
            data = buf[POS].popleft()
            if data is sentinel:
                self._sel.unregister(w_conn)
                w_conn.close()
                return
            byte = w_conn.send(data)
        except socket.error as e:
            if e.args[0] == errno.EWOULDBLOCK:
                logger.info('EWOULDBLOCK occur in send to tunnel')
                buf[POS].appendleft(data[byte:])

    def send_to_expose(self, w_conn: socket_t,
                       mask: int, buf: List[deque]) -> None:
        """send buffer data to expose"""
        if not len(buf[NEG]):
            return
        try:
            data = buf[NEG].popleft()
            if data is sentinel:
                self._sel.unregister(w_conn)
                w_conn.close()
                return
            byte = w_conn.send(data)
        except socket.error as e:
            if e.args[0] == errno.EWOULDBLOCK:
                logger.info('EWOULDBLOCK occur in send to expose')
                buf[NEG].appendleft(data[byte:])

    def _handshake(self, conn_slaver: socket_t) -> bool:
        """handshake"""
        conn_slaver.setblocking(True)  # TODO use nonblocking IO
        conn_slaver.send(self._pkgbuilder.pbuild_hs_m2s())
        buff = conn_slaver.recv(self._pkgbuilder.PACKAGE_SIZE)
        conn_slaver.setblocking(False)
        if buff == b'':  # empty response
            return False
        return self._pkgbuilder.decode_verify(buff,
                                              self._pkgbuilder.PTYPE_HS_S2M)

    def find_available_tunnel(self) -> Optional[socket_t]:
        while True:
            try:
                conn = self.tunnel_pool.popleft()
            except IndexError:
                # no available tunnel connection, just return
                # do not need to wait in a loop, because we work in LT Mode
                self.next_timeout()
                logger.info('no available tunnel connection, waiting')
                return None
            else:
                self.reset_timeout()

            if not self._handshake(conn):  # handshake first
                conn.close()
                continue
            return conn

    def run_forever(self) -> None:
        """main loop"""
        while not self._stopping:
            events = self._sel.select(timeout=self._timeout)
            for job in self._ready:  # TODO heartbeat
                job()
            self._ready.clear()

            # from pprint import pprint
            # pprint(events)
            for key, mask in events:
                callback = key.data
                callback(key.fileobj, mask)
        logger.info('stopping now ...')
        self.exit()

    def exit(self) -> None:
        """close all listening fds"""
        all_fds = chain(self._wake_fds, self.tunnel_pool,
                        self.work_pool.keys(), self._listening_sock)
        for s in all_fds:
            s.close()

    def init_wake_fds(self) -> None:
        self._wake_fds = socket.socketpair()
        for p in self._wake_fds:
            set_non_blocking(p.fileno())  # epoll need non-blocking fd

    def init_signal(self) -> None:
        self.init_wake_fds()
        signal.signal(signal.SIGINT, lambda *args: None)
        signal.set_wakeup_fd(self._wake_fds[1].fileno())
        self._sel.register(self._wake_fds[0], selectors.EVENT_READ,
                           self.handle_signal)

    def handle_signal(self, expose_sock: socket_t, mask: int) -> None:
        sig = self._wake_fds[0].recv(1)
        logger.info('recving signal {}'.format(sig))
        self._stopping = True


def parse_args() -> Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('-c', '--config', default=None, help='config path')
    parser.add_argument('-t', '--tunnel', metavar='host:port', help='')
    parser.add_argument('-b', '--bind', metavar='host:port', help='')
    parser.add_argument('-k', '--secretkey', default='secretkey', help='')
    parser.add_argument('-l', '--level', default='info', help='')
    # parser.add_argument('--ttl', default=300, type=int, dest='ttl', help='')

    return parser.parse_args()


def main() -> None:
    args = parse_args()
    config_path = args.config
    delattr(args, 'config')

    config = Config()
    if config_path is not None:
        config.load_file(config_path)
    config.from_object(args)
    try:
        config.validate([
            'tunnel',
            'bind',
            'secretkey',
        ])
    except ConfigMissing as e:
        logger.error(e)
        exit()

    logger.setLevel(name2level(config.level))

    Protocol.set_secret_key(config.secretkey)
    Protocol.recalc_crc32()
    pkgbuilder = PKGBuilder(Protocol)

    master = Server(pkgbuilder, config)
    logger.debug('PID: {}'.format(os.getpid()))
    logger.info('init successful, running as master')
    master.run_forever()


if __name__ == '__main__':
    main()



import json
from collections import UserDict
from rikka.exceptions import ConfigMissing


from typing import Callable, List, Optional, Any


class ConfigAttribute:
    """Makes an attribute forward to the config"""

    def __init__(self, name: str,
                 get_converter: Optional[Callable] = None
                 # is there a better to express callable argument
                 ) -> None:
        self.__name__ = name
        self.get_converter = get_converter

    def __get__(self, obj: Any, tp: Optional[type] = None) -> Any:
        if obj is None:
            return self
        rv = obj.config[self.__name__]
        if self.get_converter is not None:
            rv = self.get_converter(rv)
        return rv

    def __set__(self, obj: Any, value: Any) -> None:
        obj.config[self.__name__] = value


class Config(UserDict):

    def from_object(self, obj: Any) -> None:
        self.merge(obj.__dict__)

    def load_file(self, path: str) -> None:
        with open(path, 'r') as f:
            data = json.load(f)
            self.update(data)

    # def merge(self, d: Dict) -> None:
        # for k, v in d.items():
            # self.data[k] = v

    def __getattr__(self, attr: str) -> Any:
        return self.data[attr]

    def validate(self, attrs: List[str]) -> None:
        for attr in attrs:
            if self.data.get(attr) is None:
                raise ConfigMissing(attr)

    merge = UserDict.update


__all__ = [
    'Config',
    'ConfigAttribute',
]



import struct
import binascii

from typing import Any, Tuple, Sequence, Dict, Type, Optional

INTERNAL_VERSION = 0x000D
BUF_SIZE = 4096


class Protocol:

    PACKAGE_SIZE = 2 ** 6  # 64 bytes
    CTRL_PKG_TIMEOUT = 5   # Protocol recv timeout, in second

    SECRET_KEY = ''
    SECRET_KEY_CRC32 = 0
    SECRET_KEY_REVERSED_CRC32 = 0

    # Package Type
    PTYPE_HS_S2M = -1  # handshake pkg, slaver to master
    PTYPE_HEART_BEAT = 0  # heart beat pkg
    PTYPE_HS_M2S = +1  # handshake pkg, Master to Slaver

    TYPE_NAME_MAP = {
        PTYPE_HS_S2M: "PTYPE_HS_S2M",
        PTYPE_HEART_BEAT: "PTYPE_HEART_BEAT",
        PTYPE_HS_M2S: "PTYPE_HS_M2S",
    }

    FORMAT_PKG = "!b b H 20x 40s"
    FORMATS_DATA = {
        PTYPE_HS_S2M: "!I 36x",
        PTYPE_HEART_BEAT: "!40x",
        PTYPE_HS_M2S: "!I 36x",
    }

    @classmethod
    def set_secret_key(cls, secretkey: str) -> None:
        cls.SECRET_KEY = secretkey

    @classmethod
    def recalc_crc32(cls) -> None:
        cls.SECRET_KEY_CRC32 = binascii.crc32(
            cls.SECRET_KEY.encode('utf-8')
        ) & 0XFFFFFFFF
        cls.SECRET_KEY_REVERSED_CRC32 = binascii.crc32(
            cls.SECRET_KEY[::-1].encode('utf-8')
        ) & 0XFFFFFFFF


class PKGBuilder:

    # how to express subclass of Protocol
    def __init__(self,
                 protocol: Type[Protocol] = Protocol) -> None:
        self.protocol = protocol

    def __getattr__(self, name: str) -> Any:
        return getattr(self.protocol, name)

    def _build_bytes(self, pkg_ver: int = 0x01, pkg_type: int = 0,
                     prgm_ver: int = INTERNAL_VERSION,
                     data: Tuple = (), raw: Optional[bytes] = None) -> bytes:
        return struct.pack(
            self.protocol.FORMAT_PKG,
            pkg_ver,
            pkg_type,
            prgm_ver,
            raw or self.data_encode(pkg_type, data),
        )

    def data_decode(self, ptype: int, data_raw: bytes) -> Tuple:
        return struct.unpack(self.protocol.FORMATS_DATA[ptype], data_raw)

    def data_encode(self, ptype: int, data: Sequence) -> bytes:
        return struct.pack(self.protocol.FORMATS_DATA[ptype], *data)

    def pbuild_hs_m2s(self) -> bytes:
        return self._build_bytes(
            pkg_type=self.protocol.PTYPE_HS_M2S,
            data=(self.protocol.SECRET_KEY_CRC32,)
        )

    def pbuild_hs_s2m(self) -> bytes:
        return self._build_bytes(
            pkg_type=self.protocol.PTYPE_HS_S2M,
            data=(self.protocol.SECRET_KEY_REVERSED_CRC32,)
        )

    def pbuild_heart_beat(self) -> bytes:
        return self._build_bytes(pkg_type=self.PTYPE_HEART_BEAT)

    def decode_only(self, raw: bytes) -> Dict:
        if not raw or len(raw) != self.protocol.PACKAGE_SIZE:
            raise ValueError("package size should be {}, but {}".format(
                self.protocol.PACKAGE_SIZE, len(raw)
            ))
        pkg_ver, pkg_type, prgm_ver, data_raw = struct.unpack(
            self.protocol.FORMAT_PKG, raw
        )
        data = self.data_decode(pkg_type, data_raw)

        return dict(
            pkg_ver=pkg_ver,
            pkg_type=pkg_type,
            prgm_ver=prgm_ver,
            data=data,
            raw=raw,
        )

    def decode_verify(self, raw: bytes,
                      pkg_type: Optional[int] = None) -> bool:
        try:
            pkg = self.decode_only(raw)
        except ValueError:
            return False
        return self.verify(pkg=pkg)

    def verify(self, pkg: Dict) -> bool:
        pkg_type = pkg['pkg_type']
        if pkg_type == self.PTYPE_HS_S2M:
            return pkg['data'][0] == self.SECRET_KEY_REVERSED_CRC32
        elif pkg_type == self.PTYPE_HEART_BEAT:
            return True
        elif pkg_type == self.PTYPE_HS_M2S:
            return pkg['data'][0] == self.SECRET_KEY_CRC32
        return True


# use for stop
sentinel = type('Sentinel', (object,), {})



import os
import errno
import socket
import signal
import random
import selectors
import argparse
from bidict import bidict
from collections import deque
from functools import partial
from itertools import chain
from rikka.config import Config, ConfigAttribute
from rikka.exceptions import ConfigMissing
from rikka.logger import logger, name2level
from rikka.protocol import Protocol, PKGBuilder, BUF_SIZE, \
    sentinel
from rikka.utils import parse_netloc, set_non_blocking, format_addr

from argparse import Namespace
from socket import socket as socket_t
from typing import List, Optional

POS = 0  # from tunnel to dest
NEG = 1  # from dest to tunnel


class Local:

    tunnel_addr = ConfigAttribute('tunnel', parse_netloc)
    dest_addr = ConfigAttribute('dest', parse_netloc)
    max_spare_count = ConfigAttribute('max_spare_count')

    def __init__(self, pkgbuilder: PKGBuilder, config: Config) -> None:
        self._ready: deque = deque()
        self._config = config
        self._stopping = False
        self._pkgbuilder = pkgbuilder
        self._sel = selectors.DefaultSelector()

        self.tunnel_pool: deque = deque()
        self.working_pool = bidict()

        self.init_signal()
        self.reset_timeout()

    @property
    def config(self) -> Config:
        return self._config

    def next_timeout(self) -> None:
        """binary exponential backoff"""
        self._timeout_count += 1
        upper_bound = (2 ** min(self._timeout_count, 7)) - 1
        self._timeout = random.randint(1, upper_bound)

    def reset_timeout(self) -> None:
        """reset timeout to initial value"""
        self._timeout = 1
        self._timeout_count: int = 0

    def _connect_tunnel(self) -> bool:
        """establish tunnel connection"""
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        try:
            sock.connect(self.tunnel_addr)
        except ConnectionError:
            self.next_timeout()
            logger.warn(f'master is unreachable, '
                        f'retry after {self._timeout}sec...')  # noqa
            return False

        self.reset_timeout()
        self._sel.register(sock, selectors.EVENT_READ, self.prepare_transfer)
        self.tunnel_pool.append(sock)
        logger.info(f'connect to tunnel {format_addr(self.tunnel_addr)}, '
                    f'poolsize is {len(self.tunnel_pool)}')
        return True

    def _connect_dest(self) -> Optional[socket_t]:
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.settimeout(2)
        try:
            sock.connect(self.dest_addr)
        except socket.timeout:
            logger.info('connect dest timeout')
            return None
        except ConnectionRefusedError:
            logger.info('dest refused connection')
            return None

        logger.info('connected to dest {} at {}'.format(
            format_addr(sock.getpeername()),
            format_addr(sock.getsockname()),
        ))
        return sock

    def _handshake(self, conn: socket_t) -> bool:
        buff = conn.recv(self._pkgbuilder.PACKAGE_SIZE)
        if (buff == b'' or not
                self._pkgbuilder.decode_verify(
                    buff, self._pkgbuilder.PTYPE_HS_M2S
                )):
            logger.info('handshake failed')
            conn.close()
            return False

        logger.debug('handshake successful')
        conn.setblocking(True)  # TODO
        conn.send(self._pkgbuilder.pbuild_hs_s2m())
        conn.setblocking(False)
        return True

    def prepare_transfer(self, conn: socket_t, mask: int) -> None:
        self.tunnel_pool.remove(conn)
        if not self._handshake(conn):
            self._sel.unregister(conn)
            conn.close()
            return

        # connect dest after handshake successed
        sock = self._connect_dest()
        if sock is None:
            self._sel.unregister(conn)
            conn.close()
            return

        self.working_pool[conn] = sock

        buf: List[deque] = [deque(), deque()]
        self._sel.modify(conn, selectors.EVENT_WRITE | selectors.EVENT_READ,
                         partial(self.dispatch_tunnel, buf=buf))
        self._sel.register(sock, selectors.EVENT_WRITE | selectors.EVENT_READ,
                           partial(self.dispatch_dest, buf=buf))

    def dispatch_tunnel(self, conn: socket_t,
                        mask: int, buf: List[deque]) -> None:
        """schedule tunnel events"""
        if mask & selectors.EVENT_WRITE:
            self.send_to_tunnel(conn, mask, buf)
        if mask & selectors.EVENT_READ:
            self.transfer_from_tunnel(conn, mask, buf)

    def dispatch_dest(self, conn: socket_t,
                      mask: int, buf: List[deque]) -> None:
        """schedule dest events"""
        if mask & selectors.EVENT_WRITE:
            self.send_to_dest(conn, mask, buf)
        if mask & selectors.EVENT_READ:
            self.transfer_from_dest(conn, mask, buf)

    def transfer_from_tunnel(self, r_conn: socket_t,
                             mask: int, buf: List[deque]) -> None:
        """receive data from tunnel and store in buffer"""
        w_conn = self.working_pool.get(r_conn)
        if w_conn is None:
            self._sel.unregister(r_conn)
            r_conn.close()
            return

        data = b''
        need_close = False
        try:
            data = r_conn.recv(BUF_SIZE)
        except ConnectionError:
            need_close = True

        if data == b'' or need_close:
            try:
                peer = r_conn.getpeername()
                logger.info(
                    f'closing tunnel connection from {format_addr(peer)}'
                )
            except OSError as e:
                logger.warn(e)
            self._sel.unregister(r_conn)
            self._sel.modify(w_conn, selectors.EVENT_WRITE,
                             partial(self.send_to_dest, buf=buf))
            r_conn.close()
            buf[POS].append(sentinel)
            del self.working_pool[r_conn]
            return

        buf[POS].append(data)

    def transfer_from_dest(self, r_conn: socket_t,
                           mask: int, buf: List[deque]) -> None:
        """receive data from dest and store in buffer"""
        w_conn = self.working_pool.inv.get(r_conn)
        if w_conn is None:
            self._sel.unregister(r_conn)
            r_conn.close()
            return

        data = b''
        need_close = False

        try:
            data = r_conn.recv(BUF_SIZE)
        except ConnectionError:
            need_close = True

        if data == b'' or need_close:
            try:
                peer = r_conn.getpeername()
                logger.info(
                    f'closing dest connection from {format_addr(peer)}'
                )
            except OSError as e:
                logger.warn(e)
            self._sel.unregister(r_conn)
            self._sel.modify(w_conn, selectors.EVENT_WRITE,
                             partial(self.send_to_tunnel, buf=buf))
            r_conn.close()
            buf[NEG].append(sentinel)
            del self.working_pool.inv[r_conn]
            return

        buf[NEG].append(data)

    def send_to_dest(self, w_conn: socket_t,
                     mask: int, buf: List[deque]) -> None:
        if not len(buf[POS]):
            return
        try:
            data = buf[POS].popleft()
            if data is sentinel:
                self._sel.unregister(w_conn)
                w_conn.close()
                return
            byte = w_conn.send(data)
        except socket.error as e:
            if e.args[0] == errno.EWOULDBLOCK:
                logger.info('EWOULDBLOCK occur in send to dest')
                buf[POS].appendleft(data[byte:])

    def send_to_tunnel(self, w_conn: socket_t,
                       mask: int, buf: List[deque]) -> None:
        if not len(buf[NEG]):
            return
        try:
            data = buf[NEG].popleft()
            if data is sentinel:
                self._sel.unregister(w_conn)
                w_conn.close()
                return
            byte = w_conn.send(data)
        except socket.error as e:
            if e.args[0] == errno.EWOULDBLOCK:
                logger.info('EWOULDBLOCK occur in send to tunnel')
                buf[NEG].appendleft(data[byte:])

    def manage_tunnel(self) -> None:
        while len(self.tunnel_pool) < self.max_spare_count:
            if not self._connect_tunnel():  # connect failed
                break

    def run_forever(self) -> None:
        while not self._stopping:
            self.manage_tunnel()
            events = self._sel.select(timeout=self._timeout)
            for key, mask in events:
                callback = key.data
                callback(key.fileobj, mask)

        logger.info('stopping now ...')
        self.exit()

    def exit(self) -> None:
        """close all listening fds"""
        all_fds = chain(self._wake_fds, self.tunnel_pool,
                        *zip(*self.working_pool.items()))
        for s in all_fds:
            s.close()

    def init_wake_fds(self) -> None:
        self._wake_fds = socket.socketpair()
        for p in self._wake_fds:
            set_non_blocking(p.fileno())  # I found this bug from mypy check

    def init_signal(self) -> None:
        self.init_wake_fds()
        signal.signal(signal.SIGINT, lambda *args: None)
        signal.set_wakeup_fd(self._wake_fds[1].fileno())
        self._sel.register(self._wake_fds[0], selectors.EVENT_READ,
                           self.handle_signal)

    def handle_signal(self, expose_sock: socket_t, mask: int) -> None:
        sig = self._wake_fds[0].recv(1)
        logger.info('recving signal {}'.format(sig))
        self._stopping = True


def parse_args() -> Namespace:
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument('-c', '--config', default=None, help='config path')
    parser.add_argument('-t', '--tunnel', metavar='host:port', help='')
    parser.add_argument('-d', '--dest', metavar='host:port', help='')
    parser.add_argument('-k', '--secretkey', default='secretkey', help='')
    parser.add_argument('-l', '--level', default='info', help='verbose output')
    # parser.add_argument('--ttl', default=300, type=int, dest='ttl', help='')
    parser.add_argument('--max-standby', default=5, type=int,
                        dest='max_spare_count', help='')

    return parser.parse_args()


def main() -> None:
    args = parse_args()
    config_path = args.config
    delattr(args, 'config')

    config = Config()
    if config_path is not None:
        config.load_file(config_path)
    config.from_object(args)
    try:
        config.validate([
            'tunnel',
            'dest',
            'secretkey',
            'max_spare_count',
        ])
    except ConfigMissing as e:
        logger.error(e)
        exit()

    logger.setLevel(name2level(args.level))

    Protocol.set_secret_key(args.secretkey)
    Protocol.recalc_crc32()
    pkgbuilder = PKGBuilder(Protocol)

    local_ = Local(pkgbuilder, config)
    logger.debug('PID: {}'.format(os.getpid()))
    logger.info('init successful, running as local')

    local_.run_forever()


if __name__ == '__main__':
    main()



class ConfigMissing(Exception):
    """missing necessary config"""
    def __init__(self, msg: str) -> None:
        super().__init__("missing necessary config '{}'".format(msg))


class ConfigError(Exception):
    """Exception raised on config error"""


__all__ = [
    'ConfigMissing',
    'ConfigError'
]



import logging
from rikka.exceptions import ConfigError


fmt = "[%(levelname)s %(asctime)-15s] %(message)s"
timefmt = "%d %b %Y %H:%M:%S"
logger = logging.getLogger(__name__)
sh = logging.StreamHandler()
formatter = logging.Formatter(fmt, timefmt)
sh.setFormatter(formatter)
logger.addHandler(sh)


def name2level(level_name: str) -> int:
    """convert level name to logging level[int]"""
    level_name = level_name.upper()
    level = logging._nameToLevel.get(level_name)
    if level is None:
        raise ConfigError(
            "logging module doesn't support this level {}".format(level_name)
        )
    return level


__all__ = [
    'logger',
    'name2level'
]



import os
import re
import ast
from setuptools import setup


_version_re = re.compile(r'__version__\s+=\s+(.*)')
_root = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(_root, 'rikka/__init__.py')) as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read()).group(1)))

with open(os.path.join(_root, 'requirements.txt')) as f:
    requirements = f.readlines()

with open(os.path.join(_root, 'README.md')) as f:
    readme = f.read()


setup(
    name='rikka',
    version=version,
    description='Expose localhost to public Internet',
    long_description=readme,
    long_description_content_type='text/markdown',
    url='https://github.com/Hanaasagi/rikka',
    author='Hanaasagi',
    author_email='ambiguous404@gmail.com',
    license='MIT',
    python_requires='>=3.6',
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Internet',
        'Topic :: Internet :: Proxy Servers',
    ],
    packages=['rikka'],
    install_requires=requirements,
    entry_points="""
    [console_scripts]
    rklocal=rikka.local:main
    rkserver=rikka.server:main
    """,
)



import os
import socket
import fcntl


from typing import Tuple, Union
from socket import socket as socket_t


def parse_netloc(netloc: str) -> Tuple[str, int]:
    ip, _, port = netloc.rpartition(':')
    return ip, int(port)


def set_non_blocking(fd: int) -> None:
    flags = fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK
    fcntl.fcntl(fd, fcntl.F_SETFL, flags)


def format_addr(addr: Tuple) -> str:
    return '{}:{}'.format(*addr)


def create_listening_sock(addr: Union[Tuple[int], Tuple[str]]) -> socket_t:
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(addr)
    sock.listen(16)
    sock.setblocking(False)
    return sock
